"""Test some builtin potentials against galpy"""

# Third-party
import numpy as np
from astropy.constants import G
import astropy.units as u
import pytest
from scipy.special import gamma

# This project
from ...._cconfig import GSL_ENABLED
from ....units import galactic
from ..builtin import (KeplerPotential, MiyamotoNagaiPotential,
                       NFWPotential, PowerLawCutoffPotential,
                       BovyMWPotential2014)

try:
    import galpy
    import galpy.orbit
    import galpy.potential
    HAS_GALPY = True
except ImportError:
    HAS_GALPY = False

# Set to arbitrary values for testing
ro = 8.1 * u.kpc
vo = 240 * u.km/u.s
ntest = 128

def helper(gala_pot, galpy_pot):
    Rs = np.random.uniform(1, 15, size=ntest) * u.kpc
    zs = np.random.uniform(1, 15, size=ntest) * u.kpc

    xyz = np.zeros((3, Rs.size)) * u.kpc
    xyz[0] = Rs

    assert np.allclose(gala_pot.circular_velocity(xyz).to_value(u.km/u.s),
                       galpy_pot.vcirc(R=Rs.to_value(ro)))

    xyz[2] = zs
    assert np.allclose(gala_pot.density(xyz).to_value(u.Msun/u.pc**3),
                       galpy_pot.dens(R=Rs.to_value(ro), z=zs.to_value(ro)))

    assert np.allclose(gala_pot.energy(xyz).to_value((u.km / u.s)**2),
                       galpy_pot(R=Rs.to_value(ro), z=zs.to_value(ro)))

    assert np.allclose(gala_pot.gradient(xyz).to_value((u.km/u.s) * u.pc/u.Myr / u.pc)[2],
                       -galpy_pot.zforce(R=Rs.to_value(ro), z=zs.to_value(ro)))


@pytest.mark.skipif(not HAS_GALPY,
                    reason="requires galpy to run this test")
def test_kepler():
    from galpy.potential import KeplerPotential as BovyKeplerPotential

    M = 5e10 * u.Msun
    gala_pot = KeplerPotential(m=M, units=galactic)

    amp = (G*M).to_value(vo**2 * ro)
    bovy_pot = BovyKeplerPotential(amp=amp, ro=ro, vo=vo)

    helper(gala_pot, bovy_pot)


@pytest.mark.skipif(not HAS_GALPY,
                    reason="requires galpy to run this test")
def test_miyamoto():
    from galpy.potential import MiyamotoNagaiPotential as BovyMiyamotoNagaiPotential

    M = 5e10 * u.Msun
    gala_pot = MiyamotoNagaiPotential(m=M, a=3.5*u.kpc, b=300*u.pc,
                                      units=galactic)

    amp = (G*M).to_value(vo**2 * ro)
    a = gala_pot.parameters['a'].to_value(ro)
    b = gala_pot.parameters['b'].to_value(ro)
    bovy_pot = BovyMiyamotoNagaiPotential(amp=amp, a=a, b=b, ro=ro, vo=vo)

    helper(gala_pot, bovy_pot)


@pytest.mark.skipif(not HAS_GALPY,
                    reason="requires galpy to run this test")
def test_nfw():
    from galpy.potential import NFWPotential as BovyNFWPotential

    M = 6.5854e10 * u.Msun
    gala_pot = NFWPotential(m=M, r_s=15*u.kpc, units=galactic)
    amp = (G*M).to_value(vo**2 * ro)
    a = gala_pot.parameters['r_s'].to_value(ro)
    bovy_pot = BovyNFWPotential(amp, a=a, ro=ro, vo=vo)

    helper(gala_pot, bovy_pot)


@pytest.mark.skipif(not HAS_GALPY or not GSL_ENABLED,
                    reason="requires galpy and GSL to run this test")
def test_powerlawcutoff():
    from galpy.potential import PowerSphericalPotentialwCutoff

    M = 1.5854e10 * u.Msun
    alpha = 1.75
    gala_pot = PowerLawCutoffPotential(m=M, alpha=alpha, r_c=15*u.kpc,
                                       units=galactic)
    r_c = gala_pot.parameters['r_c']
    amp = (G*M).to_value(vo**2 * ro) * ((1/(2*np.pi) * r_c.to_value(ro)**(alpha - 3) / (gamma(3/2 - alpha/2))))
    bovy_pot = PowerSphericalPotentialwCutoff(amp,
                                              alpha=gala_pot.parameters['alpha'].value,
                                              rc=r_c.to_value(ro),
                                              ro=ro, vo=vo)

    helper(gala_pot, bovy_pot)


@pytest.mark.skipif(not HAS_GALPY or not GSL_ENABLED,
                    reason="requires galpy and GSL to run this test")
def test_mwpotential2014():
    from galpy.potential import (MWPotential2014,
                                 evaluateDensities,
                                 evaluatezforces,
                                 evaluatePotentials)

    # Here these have to be default:
    ro = 8 * u.kpc
    vo = 220 * u.km/u.s

    gala_pot = BovyMWPotential2014()
    bovy_pot = MWPotential2014
    for x in bovy_pot:
        x.turn_physical_on()

    Rs = np.random.uniform(1, 15, size=ntest) * u.kpc
    zs = np.random.uniform(1, 15, size=ntest) * u.kpc

    xyz = np.zeros((3, Rs.size)) * u.kpc
    xyz[0] = Rs
    xyz[2] = zs
    assert np.allclose(gala_pot.density(xyz).to_value(u.Msun/u.pc**3),
                       evaluateDensities(bovy_pot, R=Rs.to_value(ro), z=zs.to_value(ro)))

    assert np.allclose(gala_pot.energy(xyz).to_value((u.km / u.s)**2),
                       evaluatePotentials(bovy_pot, R=Rs.to_value(ro), z=zs.to_value(ro)))

    assert np.allclose(gala_pot.gradient(xyz).to_value((u.km/u.s) * u.pc/u.Myr / u.pc)[2],
                       -evaluatezforces(bovy_pot, R=Rs.to_value(ro), z=zs.to_value(ro)))



# Third-party
import astropy.coordinates as coord

# This package
from ..galactocentric import get_galactocentric2019


def test_simple():
    # Not meant to be a sensible test - that's done in Astropy!
    fr = get_galactocentric2019()
    assert isinstance(fr, coord.Galactocentric)



# Third-party
import astropy.coordinates as coord
import numpy as np
from astropy.coordinates.matrix_utilities import matrix_product

__all__ = ['transform_pm_cov']


def get_uv_tan(c):
    """Get tangent plane basis vectors on the unit sphere at the given
    spherical coordinates.
    """
    l = c.spherical.lon
    b = c.spherical.lat

    p = np.array([-np.sin(l), np.cos(l), np.zeros_like(l.value)]).T
    q = np.array([-np.cos(l)*np.sin(b), -np.sin(l)*np.sin(b), np.cos(b)]).T

    return np.stack((p, q), axis=-1)


def get_transform_matrix(from_frame, to_frame):
    """Compose sequential matrix transformations (static or dynamic) to get a
    single transformation matrix from a given path through the Astropy
    transformation machinery.

    Parameters
    ----------
    from_frame : `~astropy.coordinates.BaseCoordinateFrame` subclass
        The *class* or instance of the frame you're transforming from.
    to_frame : `~astropy.coordinates.BaseCoordinateFrame` subclass
        The class or instance of the frame you're transforming to.
    """
    if isinstance(from_frame, coord.BaseCoordinateFrame):
        from_frame_cls = from_frame.__class__
    else:
        from_frame_cls = from_frame

    if isinstance(to_frame, coord.BaseCoordinateFrame):
        to_frame_cls = to_frame.__class__
    else:
        to_frame_cls = to_frame

    path, distance = coord.frame_transform_graph.find_shortest_path(
        from_frame_cls, to_frame_cls)

    matrices = []
    currsys = from_frame
    for p in path[1:]:  # first element is fromsys so we skip it
        if isinstance(currsys, coord.BaseCoordinateFrame):
            currsys_cls = currsys.__class__
        else:
            currsys_cls = currsys
            currsys = currsys_cls()

        trans = coord.frame_transform_graph._graph[currsys_cls][p]

        if isinstance(to_frame, p):
            p = to_frame

        if isinstance(trans, coord.DynamicMatrixTransform):
            M = trans.matrix_func(currsys, p)
        elif isinstance(trans, coord.StaticMatrixTransform):
            M = trans.matrix
        else:
            raise ValueError("Transform path contains a '{0}': cannot "
                             "be composed into a single transformation "
                             "matrix.".format(trans.__class__.__name__))

        matrices.append(M)
        currsys = p

    M = None
    for Mi in reversed(matrices):
        if M is None:
            M = Mi
        else:
            M = matrix_product(M, Mi)

    return M


def transform_pm_cov(c, cov, to_frame):
    """Transform a proper motion covariance matrix to a new frame.

    Parameters
    ----------
    c : `~astropy.coordinates.SkyCoord`
        The sky coordinates of the sources in the initial coordinate frame.
    cov : array_like
        The covariance matrix of the proper motions. Must have same length as
        the input coordinates.
    to_frame : `~astropy.coordinates.BaseCoordinateFrame` subclass
        The frame to transform to as an Astropy coordinate frame class or
        instance.

    Returns
    -------
    new_cov : array_like
        The transformed covariance matrix.

    """
    if c.isscalar and cov.shape != (2, 2):
        raise ValueError('If input coordinate object is a scalar coordinate, '
                         'the proper motion covariance matrix must have shape '
                         '(2, 2), not {}'.format(cov.shape))

    elif not c.isscalar and len(c) != cov.shape[0]:
        raise ValueError('Input coordinates and covariance matrix must have '
                         'the same number of entries ({} vs {}).'
                         .format(len(c), cov.shape[0]))

    # 3D rotation matrix, to be projected onto the tangent plane
    if hasattr(c, 'frame'):
        frame = c.frame
    else:
        frame = c
    R = get_transform_matrix(frame.__class__, to_frame)

    # Get input coordinates in the desired frame:
    c_to = c.transform_to(to_frame)

    # Get tangent plane coordinates:
    uv_in = get_uv_tan(c)
    uv_to = get_uv_tan(c_to)

    if not c.isscalar:
        G = np.einsum('nab,nac->nbc', uv_to,
                      np.einsum('ji,nik->njk', R, uv_in))

        # transform
        cov_to = np.einsum('nba,nac->nbc', G,
                           np.einsum('nij,nki->njk', cov, G))
    else:
        G = np.einsum('ab,ac->bc', uv_to,
                      np.einsum('ji,ik->jk', R, uv_in))

        # transform
        cov_to = np.einsum('ba,ac->bc', G,
                           np.einsum('ij,ki->jk', cov, G))

    return cov_to



from numpy import sin,cos,deg2rad,rad2deg,arctan2,sqrt
import numpy
import numexpr

def cv_coord(a,b,c,fr=None,to=None,degr=False):
    if degr:
        degrad = deg2rad
        raddeg = rad2deg
    else:
        degrad = lambda x: x
        raddeg = lambda x: x
    if fr=='sph':
        x=c*cos(degrad(a))*cos(degrad(b))
        y=c*sin(degrad(a))*cos(degrad(b))
        z=c*sin(degrad(b))
    elif fr=='rect':
        x=a
        y=b
        z=c
    elif fr is None:
        raise Exception('You must specify the input coordinate system')
    else:
        raise Exception('Unknown input coordinate system')
    if to=='rect':
        return (x,y,z)
    elif to=='sph':
        ra = raddeg(arctan2(y,x))
        dec = raddeg(arctan2(z,sqrt(x**2+y**2)))
        rad = sqrt(x**2+y**2+z**2)
        return (ra,dec,rad)
    elif to is None:
        raise Exception('You must specify the output coordinate system')
    else:
        raise Exception('Unknown output coordinate system')

def torect(ra,dec):
    x=numexpr.evaluate('cos(ra/57.295779513082323)*cos(dec/57.295779513082323)')
    y=numexpr.evaluate('sin(ra/57.295779513082323)*cos(dec/57.295779513082323)')
    z=numexpr.evaluate('sin(dec/57.295779513082323)')
    return x,y,z

def fromrect(x,y,z):
    ra=numexpr.evaluate('arctan2(y,x)*57.295779513082323')
    dec=numexpr.evaluate('57.295779513082323*arctan2(z,sqrt(x**2+y**2))')
    return ra,dec

def sphere_rotate(ra, dec, rapol, decpol, ra0):
    """ rotate ra,dec to a new spherical coordinate system where the pole is
        at rapol,decpol and the zeropoint is at ra=ra0
        revert flag allows to reverse the transformation
    """

    x,y,z=torect(ra,dec)

    tmppol=cv_coord(rapol,decpol,1,degr=True,fr='sph',to='rect') #pole axis
    tmpvec1=cv_coord(ra0,0,1,degr=True,fr='sph',to='rect') #x axis
    tmpvec1=numpy.array(tmpvec1)

    tmpvec1[2]=(-tmppol[0]*tmpvec1[0]-tmppol[1]*tmpvec1[1])/tmppol[2]
    tmpvec1/=numpy.sqrt((tmpvec1**2).sum())
    tmpvec2=numpy.cross(tmppol,tmpvec1) # y axis

    Axx,Axy,Axz=tmpvec1
    Ayx,Ayy,Ayz=tmpvec2
    Azx,Azy,Azz=tmppol

    xnew = numexpr.evaluate('x*Axx+y*Axy+z*Axz')
    ynew = numexpr.evaluate('x*Ayx+y*Ayy+z*Ayz')
    znew = numexpr.evaluate('x*Azx+y*Azy+z*Azz')

    del x,y,z
    tmp = fromrect(xnew,ynew,znew)
    return (tmp[0],tmp[1])



# Third-party
import astropy.coordinates as coord
import astropy.units as u

__all__ = ['get_galactocentric2019']


def get_galactocentric2019():
    """
    References
    ----------
    Galactic center sky position:
    - `Reid & Brunthaler (2004) <https://ui.adsabs.harvard.edu/abs/2004ApJ...616..872R/abstract>`_

    Galactic center distance:
    - `GRAVITY collaboration (2018) <https://ui.adsabs.harvard.edu/abs/2018A%26A...615L..15G/abstract>`_

    Solar velocity:
    - `Drimmel & Poggio (2018) <https://ui.adsabs.harvard.edu/abs/2018RNAAS...2d.210D/abstract>`_
    - `Reid & Brunthaler (2004) <https://ui.adsabs.harvard.edu/abs/2004ApJ...616..872R/abstract>`_
    - `GRAVITY collaboration (2018) <https://ui.adsabs.harvard.edu/abs/2018A%26A...615L..15G/abstract>`_

    Solar height above midplane:
    - `Bennett & Bovy (2019) <https://ui.adsabs.harvard.edu/abs/2019MNRAS.482.1417B/abstract>`_

    Returns
    -------
    galcen_frame : `~astropy.coordinates.Galactocentric`
        The modern (2019) Galactocentric reference frame.

    """

    with coord.galactocentric_frame_defaults.set('v4.0'):
        galcen = coord.Galactocentric()

    return galcen



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

import os

from astropy.version import version as astropy_version

from pytest_astropy_header.display import (
    PYTEST_HEADER_MODULES,
    TESTED_VERSIONS,
    pytest_report_header as astropy_header)


def pytest_configure(config):

    config.option.astropy_header = True

    # Customize the following lines to add/remove entries from the list of
    # packages for which version numbers are displayed when running the tests.
    PYTEST_HEADER_MODULES.pop('Pandas', None)
    PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'

    from . import __version__
    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = __version__


def pytest_report_header(config):
    from gala._cconfig import GSL_ENABLED

    if GSL_ENABLED:
        hdr = " +++ Gala compiled with GSL +++"
    else:
        hdr = " --- Gala compiled without GSL ---"

    return hdr + "\n"


# Uncomment the last two lines in this block to treat all DeprecationWarnings as
# exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
# as follow (although default should work for most cases).
# To ignore some packages that produce deprecation warnings on import
# (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
# 'setuptools'), add:
#     modules_to_ignore_on_import=['module_1', 'module_2']
# To ignore some specific deprecation warning messages for Python version
# MAJOR.MINOR or later, add:
#     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# from astropy.tests.helper import enable_deprecations_as_exceptions  # noqa
# enable_deprecations_as_exceptions()
